
from pypeit.spectrographs import spectrograph
from pypeit.spectrographs import gemini_gnirs
from pypeit.spectrographs import gemini_gmos
from pypeit.spectrographs import gemini_flamingos
from pypeit.spectrographs import keck_deimos
from pypeit.spectrographs import keck_lris
from pypeit.spectrographs import keck_nires
from pypeit.spectrographs import keck_hires
from pypeit.spectrographs import keck_nirspec
from pypeit.spectrographs import keck_mosfire
from pypeit.spectrographs import magellan_fire
from pypeit.spectrographs import magellan_mage
from pypeit.spectrographs import shane_kast
from pypeit.spectrographs import tng_dolores
from pypeit.spectrographs import vlt_xshooter
from pypeit.spectrographs import wht_isis
from pypeit.spectrographs import lbt_mods
from pypeit.spectrographs import vlt_fors
from pypeit.spectrographs import lbt_luci
from pypeit.spectrographs import mmt_binospec
